import json
from tempfile import SpooledTemporaryFile
from typing import Annotated, Any, Awaitable, Callable, Dict, List, Optional

import jsonschema
from clowmdb.models import WorkflowExecution, WorkflowVersion
from fastapi import APIRouter, BackgroundTasks, Depends, HTTPException, Query, status

from app.api.dependencies import (
    AuthorizationDependency,
    CurrentUser,
    DBSession,
    HTTPClient,
    S3Service,
    get_current_workflow_execution,
    get_slurm_client,
)
from app.api.utils import (
    check_active_workflow_execution_limit,
    check_buckets_access,
    check_repo,
    start_workflow_execution,
    upload_scm_file,
)
from app.core.config import settings
from app.crud import CRUDWorkflowExecution, CRUDWorkflowVersion
from app.git_repository import GitHubRepository, build_repository
from app.schemas.workflow_execution import DevWorkflowExecutionIn, WorkflowExecutionIn, WorkflowExecutionOut
from app.scm import SCM, Provider
from app.slurm.slurm_rest_client import SlurmClient

router = APIRouter(prefix="/workflow_executions", tags=["Workflow Execution"])
workflow_authorization = AuthorizationDependency(resource="workflow_execution")

Authorization = Annotated[Callable[[str], Awaitable[Any]], Depends(workflow_authorization)]
CurrentWorkflowExecution = Annotated[WorkflowExecution, Depends(get_current_workflow_execution)]


@router.post("", status_code=status.HTTP_201_CREATED, summary="Start a new workflow execution")
async def start_workflow(
    background_tasks: BackgroundTasks,
    workflow_execution_in: WorkflowExecutionIn,
    db: DBSession,
    current_user: CurrentUser,
    authorization: Authorization,
    s3: S3Service,
    slurm_client: SlurmClient = Depends(get_slurm_client),
) -> WorkflowExecutionOut:
    """
    Start a new workflow execution. Workflow versions wit status `DEPRECATED` or `DENIED` can't be started.\n
    Permission "workflow_execution:start" required if workflow versions status is `PUBLISHED`,
    otherwise "workflow_execution:start_unpublished" required.
    \f
    Parameters
    ----------
    background_tasks : fastapi.BackgroundTasks
        Entrypoint for new BackgroundTasks. Provided by FastAPI.
    workflow_execution_in : app.schemas.workflow_executionWorkflowExecutionIn
        Meta-data and parameters for the workflow to start. HTTP Body.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    slurm_client : app.slurm.slurm_rest_client.SlurmClient
        Slurm Rest Client to communicate with Slurm cluster. Dependency Injection.

    Returns
    -------
    execution : clowmdb.models.WorkflowExecution
        Created workflow execution from the database
    """
    # Check if Workflow version exists
    workflow_version = await CRUDWorkflowVersion.get(
        db, workflow_execution_in.workflow_version_id, populate_workflow=True
    )
    if workflow_version is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Workflow version with git commit hash {workflow_execution_in.workflow_version_id} not found",
        )
    # Check authorization
    rbac_operation = "start" if workflow_version.status == WorkflowVersion.Status.PUBLISHED else "start_unpublished"
    await authorization(rbac_operation)

    if len(workflow_version.workflow_modes) > 0 and workflow_execution_in.mode is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"A workflow mode needs to be specified for thee workflow version '{workflow_execution_in.workflow_version_id}'",  # noqa: E501
        )

    # If a workflow mode is specified, check that the mode is associated with the workflow version
    workflow_mode = None
    if workflow_execution_in.mode is not None:
        workflow_mode = next(
            (mode for mode in workflow_version.workflow_modes if mode.mode_id == workflow_execution_in.mode), None
        )
        if workflow_mode is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"Workflow mode '{workflow_execution_in.mode}' does not exist on version '{workflow_execution_in.workflow_version_id}'",  # noqa: E501
            )

    # Check that the version can be used for execution
    if workflow_version.status not in [WorkflowVersion.Status.PUBLISHED, WorkflowVersion.Status.CREATED]:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=f"Workflow version with status {workflow_version.status} can't be started.",
        )
    # Check active workflow execution limit
    await check_active_workflow_execution_limit(db, current_user.uid)

    # Validate schema with saved schema in bucket
    schema_name = (
        f"{workflow_execution_in.workflow_version_id}.json"
        if workflow_execution_in.mode is None
        else f"{workflow_execution_in.workflow_version_id}-{workflow_execution_in.mode.hex}.json"
    )
    with SpooledTemporaryFile(max_size=512000) as f:
        s3.Bucket(settings.WORKFLOW_BUCKET).Object(schema_name).download_fileobj(f)
        f.seek(0)
        nextflow_schema = json.load(f)
    try:
        jsonschema.validate(workflow_execution_in.parameters, nextflow_schema)
    except jsonschema.exceptions.ValidationError as e:  # pragma: no cover
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail=f"Nextflow Parameter validation error: {e.message}"
        )

    # Create execution in database
    await check_buckets_access(
        db=db,
        parameters=workflow_execution_in.parameters,
        uid=current_user.uid,
        report_bucket=workflow_execution_in.report_output_bucket,
    )

    # Create execution in database
    execution = await CRUDWorkflowExecution.create(db, execution=workflow_execution_in, owner_id=current_user.uid)
    # Start workflow execution
    background_tasks.add_task(
        start_workflow_execution,
        s3=s3,
        db=db,
        execution=execution,
        report_output_bucket=workflow_execution_in.report_output_bucket,
        parameters=workflow_execution_in.parameters,
        git_repo=build_repository(
            url=workflow_version.workflow.repository_url, git_commit_hash=workflow_version.git_commit_hash
        ),
        slurm_client=slurm_client,
        scm_file_id=workflow_version.workflow_id.hex,
        workflow_entrypoint=workflow_mode.entrypoint if workflow_mode is not None else None,
    )

    return WorkflowExecutionOut.from_db_model(execution, workflow_id=workflow_version.workflow_id)


@router.post(
    "/arbitrary",
    status_code=status.HTTP_201_CREATED,
    summary="Start a workflow execution with arbitrary git repository",
    include_in_schema=settings.DEV_SYSTEM,
)
async def start_arbitrary_workflow(
    background_tasks: BackgroundTasks,
    workflow_execution_in: DevWorkflowExecutionIn,
    db: DBSession,
    current_user: CurrentUser,
    client: HTTPClient,
    s3: S3Service,
    authorization: Callable[[str], Awaitable[Any]] = Depends(AuthorizationDependency(resource="workflow")),
    slurm_client: SlurmClient = Depends(get_slurm_client),
) -> WorkflowExecutionOut:
    """
    Start a new workflow execution from an arbitrary git repository.\n
    Permission "workflow:create" required.\n
    For private Gitlab repositories, a Project Access Token with the role Reporter and scope `read_api` is needed.\n
    For private GitHub repositories, a Personal Access Token (classic) with scope `repo` is needed.
    \f
    Parameters
    ----------
    background_tasks : fastapi.BackgroundTasks
        Entrypoint for new BackgroundTasks. Provided by FastAPI.
    workflow_execution_in : app.schemas.workflow_executionWorkflowExecutionIn
        Meta-data and parameters for the workflow to start. HTTP Body.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    slurm_client : app.slurm.slurm_rest_client.SlurmClient
        Slurm Rest Client to communicate with Slurm cluster. Dependency Injection.
    client : httpx.AsyncClient
        Http client with an open connection. Dependency Injection.

    Returns
    -------
    execution : clowmdb.models.WorkflowExecution
        Created workflow execution from the database
    """
    if not settings.DEV_SYSTEM:  # pragma: no cover
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Not available")
    await authorization("create")
    await check_active_workflow_execution_limit(db, current_user.uid)

    try:
        # Build a git repository object based on the repository url
        repo = build_repository(
            workflow_execution_in.repository_url,
            workflow_execution_in.git_commit_hash,
            token=workflow_execution_in.token,
        )
    except NotImplementedError:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Supplied Git Repository is not supported")

    await check_repo(
        repo=repo, client=client, modes=[workflow_execution_in.mode] if workflow_execution_in.mode is not None else None
    )

    # Validate schema with saved schema in bucket
    schema_name = (
        "nextflow_schema.json" if workflow_execution_in.mode is None else workflow_execution_in.mode.schema_path
    )
    with SpooledTemporaryFile(max_size=512000) as f:
        await repo.download_file(schema_name, client=client, file_handle=f)
        f.seek(0)
        nextflow_schema = json.load(f)

    try:
        jsonschema.validate(workflow_execution_in.parameters, nextflow_schema)
    except jsonschema.exceptions.ValidationError as e:  # pragma: no cover
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail=f"Nextflow Parameter validation error: {e.message}"
        )

    # Create execution in database
    await check_buckets_access(
        db=db,
        parameters=workflow_execution_in.parameters,
        uid=current_user.uid,
        report_bucket=workflow_execution_in.report_output_bucket,
    )

    execution_note = f"Repository URL: {workflow_execution_in.repository_url}\nGit Commit Hash{workflow_execution_in.git_commit_hash}"  # noqa: E501
    if workflow_execution_in.mode is not None:
        execution_note += f"\nEntrypoint {workflow_execution_in.mode.entrypoint}"
    execution = await CRUDWorkflowExecution.create(
        db,
        execution=workflow_execution_in,
        owner_id=current_user.uid,
        notes=execution_note,
    )

    scm_provider = Provider.from_repo(repo=repo, name=f"repo{execution.execution_id.hex}")
    if repo.token is not None or not isinstance(repo, GitHubRepository):
        background_tasks.add_task(
            upload_scm_file, s3=s3, scm=SCM([scm_provider]), scm_file_id=execution.execution_id.hex
        )

    background_tasks.add_task(
        start_workflow_execution,
        s3=s3,
        db=db,
        execution=execution,
        report_output_bucket=workflow_execution_in.report_output_bucket,
        parameters=workflow_execution_in.parameters,
        git_repo=repo,
        slurm_client=slurm_client,
        scm_file_id=execution.execution_id.hex,
        workflow_entrypoint=workflow_execution_in.mode.entrypoint if workflow_execution_in.mode is not None else None,
    )

    return WorkflowExecutionOut.from_db_model(execution)


@router.get("", status_code=status.HTTP_200_OK, summary="Get all workflow executions")
async def list_workflow_executions(
    db: DBSession,
    current_user: CurrentUser,
    authorization: Authorization,
    user_id: Optional[str] = Query(
        None,
        description="Filter for workflow executions by a user. If none, Permission 'workflow_execution:read_any' required.",  # noqa: E501
        examples=["28c5353b8bb34984a8bd4169ba94c606"],
    ),
    execution_status: Optional[List[WorkflowExecution.WorkflowExecutionStatus]] = Query(
        None, description="Filter for status of workflow execution"
    ),
    workflow_version_id: Optional[str] = Query(
        None,
        description="Filter for workflow version",
        examples=["ba8bcd9294c2c96aedefa1763a84a18077c50c0f"],
        pattern=r"^[0-9a-f]{40}$",
    ),
) -> List[WorkflowExecutionOut]:
    """
    Get all workflow executions.\n
    Permission "workflow_execution:list" required, if 'user_id' is the same as the current user,
    otherwise "workflow_execution:list_all" required.
    \f
    Parameters
    ----------
    user_id : str | None, default None
        Filter for workflow executions by a user. Query Parameter.
    execution_status : List[clowmdb.models.WorkflowExecution.WorkflowExecutionStatus] | None, default None
        Filter for status of workflow execution. Query Parameter.
    workflow_version_id : str | None, default None
        Filter for workflow version, Query Parameter.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.

    Returns
    -------
    executions : List[clowmdb.models.WorkflowExecution]
        List of filtered workflow executions.
    """
    rbac_operation = "list" if user_id is not None and user_id == current_user.uid else "list_all"
    await authorization(rbac_operation)
    executions = await CRUDWorkflowExecution.list(
        db, uid=user_id, workflow_version_id=workflow_version_id, status_list=execution_status
    )
    return [
        WorkflowExecutionOut.from_db_model(
            execution, execution.workflow_version.workflow_id if execution.workflow_version is not None else None
        )
        for execution in executions
    ]


@router.get("/{eid}", status_code=status.HTTP_200_OK, summary="Get a workflow execution")
async def get_workflow_execution(
    workflow_execution: CurrentWorkflowExecution,
    current_user: CurrentUser,
    authorization: Authorization,
) -> WorkflowExecutionOut:
    """
    Get a specific workflow execution.\n
    Permission "workflow_execution:read" required if the current user started the workflow execution,
    otherwise "workflow_execution:read_any" required.
    \f
    Parameters
    ----------
    workflow_execution : clowmdb.models.WorkflowExecution
        Workflow execution with given ID. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.

    Returns
    -------
    execution : clowmdb.models.WorkflowExecution
        Workflow execution with the given id.
    """
    rbac_operation = "read" if workflow_execution.user_id == current_user.uid else "read_any"
    await authorization(rbac_operation)
    return WorkflowExecutionOut.from_db_model(
        workflow_execution,
        workflow_execution.workflow_version.workflow_id if workflow_execution.workflow_version is not None else None,
    )


@router.get("/{eid}/params", status_code=status.HTTP_200_OK, summary="Get the parameters of a workflow execution")
async def get_workflow_execution_params(
    workflow_execution: CurrentWorkflowExecution,
    current_user: CurrentUser,
    authorization: Authorization,
    s3: S3Service,
) -> Dict[str, Any]:
    """
    Get the parameters of a specific workflow execution.\n
    Permission "workflow_execution:read" required if the current user started the workflow execution,
    otherwise "workflow_execution:read_any" required.
    \f
    Parameters
    ----------
    workflow_execution : clowmdb.models.WorkflowExecution
        Workflow execution with given ID. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.

    Returns
    -------
    execution : clowmdb.models.WorkflowExecution
        Workflow execution with the given id.
    """
    rbac_operation = "read" if workflow_execution.user_id == current_user.uid else "read_any"
    await authorization(rbac_operation)
    params_file_name = f"params-{workflow_execution.execution_id.hex}.json"
    with SpooledTemporaryFile(max_size=512000) as f:
        s3.Bucket(name=settings.PARAMS_BUCKET).Object(key=params_file_name).download_fileobj(f)
        f.seek(0)
        return json.load(f)


@router.delete("/{eid}", status_code=status.HTTP_204_NO_CONTENT, summary="Delete a workflow execution")
async def delete_workflow_execution(
    background_tasks: BackgroundTasks,
    db: DBSession,
    current_user: CurrentUser,
    authorization: Authorization,
    s3: S3Service,
    workflow_execution: CurrentWorkflowExecution,
) -> None:
    """
    Delete a specific workflow execution.\n
    Permission "workflow_execution:delete" required if the current user started the workflow execution,
    otherwise "workflow_execution:delete_any" required.
    \f
    Parameters
    ----------
    background_tasks : fastapi.BackgroundTasks
        Entrypoint for new BackgroundTasks. Provided by FastAPI.
    workflow_execution : clowmdb.models.WorkflowExecution
        Workflow execution with given ID. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    """
    rbac_operation = "delete" if workflow_execution.user_id == current_user.uid else "delete_any"
    await authorization(rbac_operation)
    if workflow_execution.status in [
        WorkflowExecution.WorkflowExecutionStatus.PENDING,
        WorkflowExecution.WorkflowExecutionStatus.SCHEDULED,
        WorkflowExecution.WorkflowExecutionStatus.RUNNING,
    ]:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Cannot delete workflow execution that is not finished."
        )
    background_tasks.add_task(
        s3.Bucket(name=settings.PARAMS_BUCKET).Object(key=f"params-{workflow_execution.execution_id.hex}.json").delete
    )
    await CRUDWorkflowExecution.delete(db, workflow_execution.execution_id)


@router.post("/{eid}/cancel", status_code=status.HTTP_204_NO_CONTENT, summary="Cancel a workflow execution")
async def cancel_workflow_execution(
    background_tasks: BackgroundTasks,
    db: DBSession,
    current_user: CurrentUser,
    authorization: Authorization,
    workflow_execution: CurrentWorkflowExecution,
    slurm_client: SlurmClient = Depends(get_slurm_client),
) -> None:
    """
    Cancel a running workflow execution.\n
    Permission "workflow_execution:cancel" required if the current user started the workflow execution,
    otherwise "workflow_execution:cancel_any" required.
    \f
    Parameters
    ----------
    background_tasks : fastapi.BackgroundTasks
        Entrypoint for new BackgroundTasks. Provided by FastAPI.
    workflow_execution : clowmdb.models.WorkflowExecution
        Workflow execution with given ID. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    slurm_client : app.slurm.slurm_rest_client.SlurmClient
        Slurm Rest Client to communicate with Slurm cluster. Dependency Injection.
    """
    rbac_operation = "cancel" if workflow_execution.user_id == current_user.uid else "cancel_any"
    await authorization(rbac_operation)
    if workflow_execution.status not in [
        WorkflowExecution.WorkflowExecutionStatus.PENDING,
        WorkflowExecution.WorkflowExecutionStatus.SCHEDULED,
        WorkflowExecution.WorkflowExecutionStatus.RUNNING,
    ]:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Cannot cancel workflow execution that is finished."
        )
    if workflow_execution.slurm_job_id >= 0:
        background_tasks.add_task(slurm_client.cancel_job, job_id=workflow_execution.slurm_job_id)
    await CRUDWorkflowExecution.cancel(db, workflow_execution.execution_id)
