from enum import Enum, unique
from typing import Annotated, Any, Awaitable, Callable, List, Optional
from uuid import UUID

from clowmdb.models import WorkflowVersion
from fastapi import APIRouter, BackgroundTasks, Depends, File, HTTPException, Path, Query, UploadFile, status
from fastapi.responses import StreamingResponse

from app.api.dependencies import (
    AuthorizationDependency,
    CurrentUser,
    CurrentWorkflow,
    CurrentWorkflowVersion,
    DBSession,
    HTTPClient,
    S3Service,
)
from app.api.utils import delete_remote_icon, upload_icon
from app.core.config import settings
from app.crud import CRUDWorkflowVersion
from app.git_repository import build_repository
from app.schemas.workflow_version import IconUpdateOut
from app.schemas.workflow_version import WorkflowVersion as WorkflowVersionSchema
from app.schemas.workflow_version import WorkflowVersionStatus

router = APIRouter(prefix="/{wid}/versions", tags=["Workflow Version"])
workflow_authorization = AuthorizationDependency(resource="workflow")

Authorization = Annotated[Callable[[str], Awaitable[Any]], Depends(workflow_authorization)]


@unique
class DocumentationEnum(Enum):
    USAGE: str = "usage"
    INPUT: str = "input"
    OUTPUT: str = "output"
    CHANGELOG: str = "changelog"
    PARAMETER_SCHEMA: str = "parameter_schema"

    def standard_path(self) -> str:
        if self is DocumentationEnum.USAGE:
            return "README.md"
        elif self is DocumentationEnum.INPUT:
            return "docs/usage.md"
        elif self is DocumentationEnum.OUTPUT:
            return "docs/output.md"
        elif self is DocumentationEnum.CHANGELOG:
            return "CHANGELOG.md"
        return "nextflow_schema.json"


@router.get("", status_code=status.HTTP_200_OK, summary="Get all versions of a workflow")
async def list_workflow_version(
    current_user: CurrentUser,
    workflow: CurrentWorkflow,
    db: DBSession,
    authorization: Authorization,
    version_status: Optional[List[WorkflowVersion.Status]] = Query(
        None,
        description=f"Which versions of the workflow to include in the response. Permission 'workflow:list_filter' required if you are not the developer of this workflow. Default {WorkflowVersion.Status.PUBLISHED.name} and {WorkflowVersion.Status.DEPRECATED.name}",  # noqa: E501
    ),
) -> List[WorkflowVersionSchema]:
    """
    List all versions of a Workflow.\n
    Permission "workflow:list" required.
    \f
    Parameters
    ----------
    workflow : clowmdb.models.Workflow
        Workflow with given ID. Dependency Injection.
    version_status : List[clowmdb.models.WorkflowVersion.Status] | None, default None
        Status of Workflow versions to filter for to fetch. Query Parameter
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.

    Returns
    -------
    versions : [app.schemas.workflow_version.WorkflowVersion]
        All versions of the workflow
    """
    rbac_operation = (
        "list_filter" if workflow.developer_id != current_user.uid and version_status is not None else "list"
    )
    await authorization(rbac_operation)

    versions = await CRUDWorkflowVersion.list(
        db,
        workflow.workflow_id,
        version_status=version_status
        if version_status is not None
        else [WorkflowVersion.Status.PUBLISHED, WorkflowVersion.Status.DEPRECATED],
    )
    return [WorkflowVersionSchema.from_db_version(v, load_modes=True) for v in versions]


@router.get(
    "/{git_commit_hash}",
    status_code=status.HTTP_200_OK,
    summary="Get a workflow version",
)
async def get_workflow_version(
    workflow: CurrentWorkflow,
    db: DBSession,
    current_user: CurrentUser,
    authorization: Authorization,
    git_commit_hash: str = Path(
        ...,
        description="Git commit git_commit_hash of specific version or 'latest'.",
        pattern=r"^([0-9a-f]{40}|latest)$",
        examples=["latest", "ba8bcd9294c2c96aedefa1763a84a18077c50c0f"],
    ),
) -> WorkflowVersionSchema:
    """
    Get a specific version of a workflow.\n
    Permission "workflow:read" required if the version is public or you are the developer of the workflow,
    otherwise "workflow:read_any"
    \f
    Parameters
    ----------
    workflow : clowmdb.models.Workflow
        Workflow with given ID. Dependency Injection.
    git_commit_hash: str
        Version ID
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.

    Returns
    -------
    version : app.schemas.workflow_version.WorkflowVersion
        The specified WorkflowVersion
    """
    rbac_operation = "read"
    version = (
        await CRUDWorkflowVersion.get_latest(db, workflow.workflow_id)
        if git_commit_hash == "latest"
        else await CRUDWorkflowVersion.get(db, git_commit_hash, workflow_id=workflow.workflow_id)
    )
    if version is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Workflow Version with git_commit_hash '{git_commit_hash}' not found",
        )
    # Developer can read any version of his workflow, others only published and deprecated ones
    if (
        current_user.uid != workflow.developer_id
        and version.status != WorkflowVersion.Status.PUBLISHED
        and version.status != WorkflowVersion.Status.DEPRECATED
    ):
        rbac_operation = "read_any"
    await authorization(rbac_operation)
    return WorkflowVersionSchema.from_db_version(version, load_modes=True)


@router.patch("/{git_commit_hash}/status", status_code=status.HTTP_200_OK, summary="Update status of workflow version")
async def update_workflow_version_status(
    version_status: WorkflowVersionStatus,
    workflow_version: CurrentWorkflowVersion,
    db: DBSession,
    authorization: Authorization,
) -> WorkflowVersionSchema:
    """
    Update the status of a workflow version.\n
    Permission "workflow:update_status"
    \f
    Parameters
    ----------
    version_status : app.schemas.workflow_version.WorkflowVersionStatus
        New Status of the workflow version. HTTP Body.
    workflow_version : clowmdb.models.WorkflowVersion
        Workflow version with given ID. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.

    Returns
    -------
    version : clowmdb.models.WorkflowVersion
        Version of the workflow with updated status
    """
    await authorization("update_status")
    await CRUDWorkflowVersion.update_status(db, workflow_version.git_commit_hash, version_status.status)
    workflow_version.status = version_status.status
    return WorkflowVersionSchema.from_db_version(workflow_version, load_modes=True)


@router.post("/{git_commit_hash}/deprecate", status_code=status.HTTP_200_OK, summary="Deprecate a workflow version")
async def deprecate_workflow_version(
    workflow: CurrentWorkflow,
    workflow_version: CurrentWorkflowVersion,
    db: DBSession,
    authorization: Authorization,
    current_user: CurrentUser,
) -> WorkflowVersionSchema:
    """
    Deprecate a workflow version.\n
    Permission "workflow:update" required if you are the developer of the workflow,
    otherwise "workflow:read_status"
    \f
    Parameters
    ----------
    workflow : clowmdb.models.Workflow
        Workflow with given ID. Dependency Injection.
    workflow_version : clowmdb.models.WorkflowVersion
        Workflow version with given ID. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    current_user: clowmdb.models.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.

    Returns
    -------
    version : clowmdb.models.WorkflowVersion
        Version of the workflow with deprecated status
    """
    await authorization("update_status" if current_user.uid != workflow.developer_id else "update")
    await CRUDWorkflowVersion.update_status(db, workflow_version.git_commit_hash, WorkflowVersion.Status.DEPRECATED)
    workflow_version.status = WorkflowVersion.Status.DEPRECATED
    return WorkflowVersionSchema.from_db_version(workflow_version, load_modes=True)


@router.get(
    "/{git_commit_hash}/documentation",
    status_code=status.HTTP_200_OK,
    summary="Fetch documentation for a workflow version",
    response_class=StreamingResponse,
)
async def download_workflow_documentation(
    workflow: CurrentWorkflow,
    workflow_version: CurrentWorkflowVersion,
    authorization: Authorization,
    client: HTTPClient,
    db: DBSession,
    document: DocumentationEnum = Query(
        DocumentationEnum.USAGE, description="Specific which type of documentation the client wants to fetch"
    ),
    mode_id: Optional[UUID] = Query(default=None, description="Workflow Mode"),
) -> StreamingResponse:
    """
    Get the documentation for a specific workflow version.
    Streams the response directly from the right git repository.\n
    Permission "workflow:read" required.
    \f
    Parameters
    ----------
    workflow : clowmdb.models.Workflow
        Workflow with given ID. Dependency Injection.
    workflow_version : clowmdb.models.WorkflowVersion
        Workflow version with given ID. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    client : httpx.AsyncClient
        HTTP Client with an open connection. Dependency Injection.
    document : DocumentationEnum, default DocumentationEnum.USAGE
        Which type of documentation the client wants to fetch
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    mode_id : UUID | None
        Select the workflow mode of the workflow version

    Returns
    -------
    response : StreamingResponse
        Streams the requested document from the git repository directly to the client
    """
    await authorization("read")
    repo = build_repository(
        workflow.repository_url,
        workflow_version.git_commit_hash,
        workflow.credentials_token,
    )
    path = document.standard_path()
    if mode_id is not None:
        workflow_mode = next((mode for mode in workflow_version.workflow_modes if mode.mode_id == mode_id), None)
        if workflow_mode is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=f"Workflow mode with ID '{mode_id}' not found"
            )
        if document is DocumentationEnum.PARAMETER_SCHEMA:
            path = workflow_mode.schema_path

    return StreamingResponse(repo.download_file_stream(path, client), headers={"Cache-Control": "max-age=86400"})


@router.post(
    "/{git_commit_hash}/icon",
    status_code=status.HTTP_201_CREATED,
    summary="Upload icon for workflow version",
)
async def upload_workflow_version_icon(
    workflow: CurrentWorkflow,
    background_tasks: BackgroundTasks,
    workflow_version: CurrentWorkflowVersion,
    s3: S3Service,
    authorization: Authorization,
    current_user: CurrentUser,
    db: DBSession,
    icon: UploadFile = File(..., description="Optional Icon for the Workflow."),
) -> IconUpdateOut:
    """
    Upload an icon for the workflow version and returns the new icon URL.\n
    Permission "workflow:update" required.
    \f
    Parameters
    ----------
    workflow : clowmdb.models.Workflow
        Workflow with given ID. Dependency Injection.
    background_tasks : fastapi.BackgroundTasks
        Entrypoint for new BackgroundTasks. Provided by FastAPI.
    workflow_version : clowmdb.models.WorkflowVersion
        Workflow version with given ID. Dependency Injection.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    current_user : clowmdb.models.User
        Current user. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    icon : fastapi.UploadFile
        New Icon for the workflow version. HTML Form.

    Returns
    -------
    icon_url : str
        URL where the icon can be downloaded
    """
    await authorization("update")
    if current_user.uid != workflow.developer_id:
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="Only the developer can update his workflow")
    old_slug = workflow_version.icon_slug
    icon_slug = upload_icon(s3=s3, background_tasks=background_tasks, icon=icon)
    await CRUDWorkflowVersion.update_icon(db, workflow_version.git_commit_hash, icon_slug)
    # Delete old icon if possible
    if old_slug is not None:
        background_tasks.add_task(delete_remote_icon, s3=s3, db=db, icon_slug=old_slug)
    return IconUpdateOut(icon_url=str(settings.OBJECT_GATEWAY_URI) + "/".join([settings.ICON_BUCKET, icon_slug]))


@router.delete(
    "/{git_commit_hash}/icon",
    status_code=status.HTTP_204_NO_CONTENT,
    summary="Delete icon of workflow version",
)
async def delete_workflow_version_icon(
    workflow: CurrentWorkflow,
    workflow_version: CurrentWorkflowVersion,
    background_tasks: BackgroundTasks,
    authorization: Authorization,
    current_user: CurrentUser,
    s3: S3Service,
    db: DBSession,
) -> None:
    """
    Delete the icon of the workflow version.\n
    Permission "workflow:update" required.
    \f
    Parameters
    ----------
    workflow : clowmdb.models.Workflow
        Workflow with given ID. Dependency Injection.
    workflow_version : clowmdb.models.WorkflowVersion
        Workflow version with given ID. Dependency Injection.
    background_tasks : fastapi.BackgroundTasks
        Entrypoint for new BackgroundTasks. Provided by FastAPI.
    authorization : Callable[[str], Awaitable[Any]]
        Async function to ask the auth service for authorization. Dependency Injection.
    current_user : clowmdb.models.User
        Current user. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    """
    await authorization("update")
    if current_user.uid != workflow.developer_id:
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="Only the developer can update his workflow")
    if workflow_version.icon_slug is not None:
        background_tasks.add_task(delete_remote_icon, s3=s3, db=db, icon_slug=workflow_version.icon_slug)
        await CRUDWorkflowVersion.update_icon(db, workflow_version.git_commit_hash, icon_slug=None)
