from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import AnyHttpUrl, AnyUrl, Field, computed_field
from pydantic_settings import BaseSettings, SettingsConfigDict


def _assemble_db_uri(values: Dict[str, Any], async_flag: bool = True) -> Any:
    return AnyUrl.build(
        scheme=f"mysql+{'aiomysql' if async_flag else 'pymysql'}",
        password=values.get("DB_PASSWORD"),
        username=values.get("DB_USER"),
        port=values.get("DB_PORT"),
        host=values.get("DB_HOST"),  # type: ignore[arg-type]
        path=f"{values.get('DB_DATABASE') or ''}",
    )


def _load_public_key(pub_key_val: Optional[str], pub_key_file: Optional[Path]) -> str:
    pub_key = ""
    if pub_key_val is not None:
        pub_key = pub_key_val
    if pub_key_file is not None:
        with open(pub_key_file) as f:
            pub_key = f.read()
    if len(pub_key) == 0:
        raise ValueError("PUBLIC_KEY_VALUE or PUBLIC_KEY_FILE must be set")
    return pub_key


class Settings(BaseSettings):
    API_PREFIX: str = Field("/api/workflow-service", description="Path Prefix for all API endpoints.")

    public_key_value: Optional[str] = Field(
        None, description="Public RSA Key in PEM format to sign the JWTs.", validation_alias="PUBLIC_KEY_VALUE"
    )
    public_key_file: Optional[Path] = Field(
        None, description="Path to Public RSA Key in PEM format to sign the JWTs.", validation_alias="PUBLIC_KEY_FILE"
    )

    @computed_field  # type: ignore[misc]
    @property
    def PUBLIC_KEY(self) -> str:
        return _load_public_key(self.public_key_value, self.public_key_file)

    # BACKEND_CORS_ORIGINS is a JSON-formatted list of origins
    BACKEND_CORS_ORIGINS: List[AnyHttpUrl] = Field([], description="List of all valid CORS origins")

    DB_HOST: str = Field(..., description="Host of the database.")
    DB_USER: str = Field(..., description="Username in the database.")
    DB_PASSWORD: str = Field(..., description="Password for the database user.")
    DB_DATABASE: str = Field(..., description="Name of the database.")
    DB_PORT: int = Field(3306, description="Port of the database.")
    SQLALCHEMY_VERBOSE_LOGGER: bool = Field(False, description="Flag whether to print the SQL Queries in the logs.")

    @computed_field  # type: ignore[misc]
    @property
    def SQLALCHEMY_DATABASE_ASYNC_URI(self) -> AnyUrl:
        return _assemble_db_uri(
            {
                "DB_HOST": self.DB_HOST,
                "DB_USER": self.DB_USER,
                "DB_PASSWORD": self.DB_PASSWORD,
                "DB_DATABASE": self.DB_DATABASE,
                "DB_PORT": self.DB_PORT,
            },
            async_flag=True,
        )

    @computed_field  # type: ignore[misc]
    @property
    def SQLALCHEMY_DATABASE_NORMAL_URI(self) -> AnyUrl:
        return _assemble_db_uri(
            {
                "DB_HOST": self.DB_HOST,
                "DB_USER": self.DB_USER,
                "DB_PASSWORD": self.DB_PASSWORD,
                "DB_DATABASE": self.DB_DATABASE,
                "DB_PORT": self.DB_PORT,
            },
            async_flag=False,
        )

    OBJECT_GATEWAY_URI: AnyHttpUrl = Field(..., description="URI of the Ceph Object Gateway.")
    BUCKET_CEPH_ACCESS_KEY: str = Field(
        ..., description="Access key for the Ceph Object Gateway with admin privileges."
    )
    BUCKET_CEPH_SECRET_KEY: str = Field(
        ..., description="Secret key for the Ceph Object Gateway with admin privileges."
    )
    OPA_URI: AnyHttpUrl = Field(..., description="URI of the OPA Service")
    OPA_POLICY_PATH: str = Field("/clowm/authz/allow", description="Path to the OPA Policy for Authorization")
    PARAMS_BUCKET: str = Field("nxf-params", description="Bucket where the nextflow configurations should be saved")
    WORKFLOW_BUCKET: str = Field("clowm-workflows", description="Bucket where to save workflow files")
    ICON_BUCKET: str = Field("clowm-icons", description="Bucket where to save workflow icons")
    SLURM_ENDPOINT: AnyHttpUrl = Field(..., description="URI of the Slurm Cluster")
    SLURM_TOKEN: str = Field(..., description="JWT for authentication the Slurm REST API")
    SLURM_USER: str = Field("slurm", description="User of communication with slurm")
    PARAMS_BUCKET_MOUNT_PATH: str = Field(
        "/mnt/params-bucket", description="Path on the slurm cluster where the params bucket is mounted."
    )
    NX_CONFIG: Optional[str] = Field(None, description="Path to a nextflow configuration for every run")
    NX_BIN: str = Field("nextflow", description="Path to the nextflow executable")
    SLURM_WORKING_DIRECTORY: str = Field(
        "/tmp", description="Working directory for the slurm job with the nextflow command"
    )
    ACTIVE_WORKFLOW_EXECUTION_LIMIT: int = Field(3, description="The limit of active workflow executions per user.")
    SLURM_JOB_STATUS_CHECK_INTERVAL: int = Field(
        30,
        ge=1,
        le=600,
        description="Interval for checking the slurm jobs status after starting a workflow execution in seconds",
    )
    DEV_SYSTEM: bool = Field(False, description="Open a endpoint where to execute arbitrary workflows.")
    model_config = SettingsConfigDict(case_sensitive=True, env_file=".env", secrets_dir="/run/secrets", extra="ignore")


settings = Settings()
